@@ -39,34 +39,35 @@
 /*
   Include declarations.
 */
-#include "magick/studio.h"
-#include "magick/artifact.h"
-#include "magick/blob.h"
-#include "magick/blob-private.h"
-#include "magick/client.h"
-#include "magick/colorspace-private.h"
-#include "magick/display.h"
-#include "magick/exception.h"
-#include "magick/exception-private.h"
-#include "magick/image.h"
-#include "magick/image-private.h"
-#include "magick/list.h"
-#include "magick/magick.h"
-#include "magick/monitor.h"
-#include "magick/monitor-private.h"
-#include "magick/memory_.h"
-#include "magick/option.h"
-#include "magick/pixel-accessor.h"
-#include "magick/profile.h"
-#include "magick/property.h"
-#include "magick/quantum-private.h"
-#include "magick/static.h"
-#include "magick/string_.h"
-#include "magick/string-private.h"
-#include "magick/module.h"
-#include "magick/utility.h"
-#include "magick/xwindow.h"
-#include "magick/xwindow-private.h"
+#include "MagickCore/studio.h"
+#include "MagickCore/artifact.h"
+#include "MagickCore/blob.h"
+#include "MagickCore/blob-private.h"
+#include "MagickCore/client.h"
+#include "MagickCore/colorspace-private.h"
+#include "MagickCore/display.h"
+#include "MagickCore/exception.h"
+#include "MagickCore/exception-private.h"
+#include "MagickCore/image.h"
+#include "MagickCore/image-private.h"
+#include "MagickCore/layer.h"
+#include "MagickCore/list.h"
+#include "MagickCore/magick.h"
+#include "MagickCore/monitor.h"
+#include "MagickCore/monitor-private.h"
+#include "MagickCore/memory_.h"
+#include "MagickCore/option.h"
+#include "MagickCore/pixel-accessor.h"
+#include "MagickCore/profile.h"
+#include "MagickCore/property.h"
+#include "MagickCore/quantum-private.h"
+#include "MagickCore/static.h"
+#include "MagickCore/string_.h"
+#include "MagickCore/string-private.h"
+#include "MagickCore/module.h"
+#include "MagickCore/utility.h"
+#include "MagickCore/xwindow.h"
+#include "MagickCore/xwindow-private.h"
 #if defined(MAGICKCORE_WEBP_DELEGATE)
 #include <webp/decode.h>
 #include <webp/encode.h>
@@ -81,7 +82,7 @@
 */
 #if defined(MAGICKCORE_WEBP_DELEGATE)
 static MagickBooleanType
-  WriteWEBPImage(const ImageInfo *,Image *);
+  WriteWEBPImage(const ImageInfo *,Image *,ExceptionInfo *);
 #endif
 
 /*
@@ -226,7 +227,8 @@ static int FillBasicWEBPInfo(Image *image,const uint8_t *stream,size_t length,
   image->columns=(size_t) features->width;
   image->rows=(size_t) features->height;
   image->depth=8;
-  image->matte=features->has_alpha != 0 ? MagickTrue : MagickFalse;
+  image->alpha_trait=features->has_alpha != 0 ? BlendPixelTrait :
+        UndefinedPixelTrait;
 
   return(webp_status);
 }
@@ -294,33 +296,33 @@ static int ReadSingleWEBPImage(Image *image,const uint8_t *stream,
   p=(unsigned char *) webp_image->u.RGBA.rgba;
   for (y=0; y < (ssize_t) image->rows; y++)
   {
-    register PixelPacket
+    register Quantum
       *q;
 
     register ssize_t
       x;
 
     q=QueueAuthenticPixels(image,0,y,image->columns,1,exception);
-    if (q == (PixelPacket *) NULL)
+    if (q == (Quantum *) NULL)
       break;
     for (x=0; x < (ssize_t) image->columns; x++)
     {
       if ((x >= x_offset && x < x_offset + image_width) &&
           (y >= y_offset && y < y_offset + image_height))
         {
-          SetPixelRed(q,ScaleCharToQuantum(*p++));
-          SetPixelGreen(q,ScaleCharToQuantum(*p++));
-          SetPixelBlue(q,ScaleCharToQuantum(*p++));
-          SetPixelAlpha(q,ScaleCharToQuantum(*p++));
+          SetPixelRed(image,ScaleCharToQuantum(*p++),q);
+          SetPixelGreen(image,ScaleCharToQuantum(*p++),q);
+          SetPixelBlue(image,ScaleCharToQuantum(*p++),q);
+          SetPixelAlpha(image,ScaleCharToQuantum(*p++),q);
         }
       else
         {
-          SetPixelRed(q,0);
-          SetPixelGreen(q,0);
-          SetPixelBlue(q,0);
-          SetPixelAlpha(q,0);
+          SetPixelRed(image,0,q);
+          SetPixelGreen(image,0,q);
+          SetPixelBlue(image,0,q);
+          SetPixelAlpha(image,0,q);
         }
-      q++;
+      q+=GetPixelChannels(image);
     }
     if (SyncAuthenticPixels(image,exception) == MagickFalse)
       break;
@@ -346,7 +348,8 @@ static int ReadSingleWEBPImage(Image *image,const uint8_t *stream,
       *mux;
 
     /*
-      Extract any profiles.
+      Extract any profiles:
+      https://developers.google.com/speed/webp/docs/container-api.
     */
     content.bytes=stream;
     content.size=length;
@@ -359,7 +362,7 @@ static int ReadSingleWEBPImage(Image *image,const uint8_t *stream,
         profile=BlobToStringInfo(chunk.bytes,chunk.size);
         if (profile != (StringInfo *) NULL)
           {
-            SetImageProfile(image,"ICC",profile);
+            SetImageProfile(image,"ICC",profile,exception);
             profile=DestroyStringInfo(profile);
           }
       }
@@ -369,7 +372,7 @@ static int ReadSingleWEBPImage(Image *image,const uint8_t *stream,
         profile=BlobToStringInfo(chunk.bytes,chunk.size);
         if (profile != (StringInfo *) NULL)
           {
-            SetImageProfile(image,"EXIF",profile);
+            SetImageProfile(image,"EXIF",profile,exception);
             profile=DestroyStringInfo(profile);
           }
       }
@@ -379,7 +382,7 @@ static int ReadSingleWEBPImage(Image *image,const uint8_t *stream,
         profile=BlobToStringInfo(chunk.bytes,chunk.size);
         if (profile != (StringInfo *) NULL)
           {
-            SetImageProfile(image,"XMP",profile);
+            SetImageProfile(image,"XMP",profile,exception);
             profile=DestroyStringInfo(profile);
           }
       }
@@ -423,41 +426,47 @@ static int ReadAnimatedWEBPImage(const ImageInfo *image_info,Image *image,
   data.bytes=stream;
   data.size=length;
   demux=WebPDemux(&data);
-  if (WebPDemuxGetFrame(demux,1,&iter)) {
-    do {
-      if (image_count != 0)
-        {
-          AcquireNextImage(image_info,image);
-          if (GetNextImageInList(image) == (Image *) NULL)
-            break;
-          image=SyncNextImageInList(image);
-          CloneImageProperties(image, original_image);
-          image->page.x=iter.x_offset;
-          image->page.y=iter.y_offset;
-          webp_status=ReadSingleWEBPImage(image,iter.fragment.bytes,
-            iter.fragment.size,configure,exception,MagickFalse);
-        }
-      else
-        {
-          image->page.x=iter.x_offset;
-          image->page.y=iter.y_offset;
-          webp_status=ReadSingleWEBPImage(image,iter.fragment.bytes,
-            iter.fragment.size,configure,exception,MagickTrue);
-        }
-      if (webp_status != VP8_STATUS_OK)
-        break;
-
-      image->page.width=canvas_width;
-      image->page.height=canvas_height;
-      image->ticks_per_second=100;
-      image->delay=iter.duration/10;
-      image->dispose=NoneDispose;
-      if (iter.dispose_method == WEBP_MUX_DISPOSE_BACKGROUND)
-        image->dispose=BackgroundDispose;
-      image_count++;
-    } while (WebPDemuxNextFrame(&iter));
-    WebPDemuxReleaseIterator(&iter);
-  }
+  if (WebPDemuxGetFrame(demux,1,&iter))
+    {
+      do
+      {
+        if (image_count != 0)
+          {
+            AcquireNextImage(image_info,image,exception);
+            if (GetNextImageInList(image) == (Image *) NULL)
+              break;
+            image=SyncNextImageInList(image);
+            CloneImageProperties(image,original_image);
+            image->page.x=iter.x_offset;
+            image->page.y=iter.y_offset;
+            webp_status=ReadSingleWEBPImage(image,iter.fragment.bytes,
+              iter.fragment.size,configure,exception,MagickFalse);
+          }
+        else
+          {
+            image->page.x=iter.x_offset;
+            image->page.y=iter.y_offset;
+            webp_status=ReadSingleWEBPImage(image,iter.fragment.bytes,
+              iter.fragment.size,configure,exception,MagickTrue);
+          }
+        if (webp_status != VP8_STATUS_OK)
+          break;
+        image->page.width=canvas_width;
+        image->page.height=canvas_height;
+        image->ticks_per_second=100;
+        image->delay=iter.duration/10;
+        image->dispose=NoneDispose;
+        if (iter.dispose_method == WEBP_MUX_DISPOSE_BACKGROUND)
+          image->dispose=BackgroundDispose;
+        (void) SetImageProperty(image,"webp:mux-blend",
+          "AtopPreviousAlphaBlend",exception);
+        if (iter.blend_method == WEBP_MUX_BLEND)
+          (void) SetImageProperty(image,"webp:mux-blend",
+            "AtopBackgroundAlphaBlend",exception);
+        image_count++;
+      } while (WebPDemuxNextFrame(&iter));
+      WebPDemuxReleaseIterator(&iter);
+    }
   WebPDemuxDelete(demux);
   return(webp_status);
 }
@@ -510,7 +519,7 @@ static Image *ReadWEBPImage(const ImageInfo *image_info,
       image_info->filename);
   assert(exception != (ExceptionInfo *) NULL);
   assert(exception->signature == MagickCoreSignature);
-  image=AcquireImage(image_info);
+  image=AcquireImage(image_info,exception);
   status=OpenBlob(image_info,image,ReadBinaryBlobMode,exception);
   if (status == MagickFalse)
     {
@@ -634,7 +643,7 @@ ModuleExport size_t RegisterWEBPImage(void)
     *entry;
 
   *version='\0';
-  entry=SetMagickInfo("WEBP");
+  entry=AcquireMagickInfo("WEBP","WEBP","WebP Image Format");
 #if defined(MAGICKCORE_WEBP_DELEGATE)
   entry->decoder=(DecodeImageHandler *) ReadWEBPImage;
   entry->encoder=(EncodeImageHandler *) WriteWEBPImage;
@@ -643,13 +652,11 @@ ModuleExport size_t RegisterWEBPImage(void)
     (WebPGetEncoderVersion() >> 8) & 0xff,
     (WebPGetEncoderVersion() >> 0) & 0xff,WEBP_ENCODER_ABI_VERSION);
 #endif
-  entry->description=ConstantString("WebP Image Format");
   entry->mime_type=ConstantString("image/webp");
-  entry->seekable_stream=MagickTrue;
+  entry->flags|=CoderDecoderSeekableStreamFlag;
 #if !defined(MAGICKCORE_WEBPMUX_DELEGATE)
-  entry->adjoin=MagickFalse;
+  entry->flags^=CoderAdjoinFlag;
 #endif
-  entry->magick_module=ConstantString("WEBP");
   entry->magick=(IsImageFormatHandler *) IsWEBP;
   if (*version != '\0')
     entry->version=ConstantString(version);
@@ -767,7 +774,7 @@ static MagickBooleanType WriteSingleWEBPImage(const ImageInfo *image_info,
   /*
     Allocate memory for pixels.
   */
-  (void) TransformImageColorspace(image,sRGBColorspace);
+  (void) TransformImageColorspace(image,sRGBColorspace,exception);
   picture_memory->pixel_info=AcquireVirtualMemory(image->columns,image->rows*
     sizeof(*(picture->argb)));
 
@@ -780,23 +787,23 @@ static MagickBooleanType WriteSingleWEBPImage(const ImageInfo *image_info,
   q=picture->argb;
   for (y=0; y < (ssize_t) image->rows; y++)
   {
-    register const PixelPacket
+    register const Quantum
       *magick_restrict p;
 
     register ssize_t
       x;
 
     p=GetVirtualPixels(image,0,y,image->columns,1,exception);
-    if (p == (const PixelPacket *) NULL)
+    if (p == (const Quantum *) NULL)
       break;
     for (x=0; x < (ssize_t) image->columns; x++)
     {
-      *q++=(uint32_t) (image->matte != MagickFalse ? (uint32_t)
-        ScaleQuantumToChar(GetPixelAlpha(p)) << 24 : 0xff000000) |
-        ((uint32_t) ScaleQuantumToChar(GetPixelRed(p)) << 16) |
-        ((uint32_t) ScaleQuantumToChar(GetPixelGreen(p)) << 8) |
-        ((uint32_t) ScaleQuantumToChar(GetPixelBlue(p)));
-      p++;
+      *q++=(uint32_t) (image->alpha_trait != UndefinedPixelTrait ? (uint32_t)
+        ScaleQuantumToChar(GetPixelAlpha(image,p)) << 24 : 0xff000000) |
+        ((uint32_t) ScaleQuantumToChar(GetPixelRed(image,p)) << 16) |
+        ((uint32_t) ScaleQuantumToChar(GetPixelGreen(image,p)) << 8) |
+        ((uint32_t) ScaleQuantumToChar(GetPixelBlue(image,p)));
+      p+=GetPixelChannels(image);
     }
     status=SetImageProgress(image,SaveImageTag,(MagickOffsetType) y,
       image->rows);
@@ -889,7 +896,7 @@ static MagickBooleanType WriteAnimatedWEBPImage(const ImageInfo *image_info,
 #endif
 
 static MagickBooleanType WriteWEBPImage(const ImageInfo *image_info,
-  Image *image)
+  Image *image,ExceptionInfo * exception)
 {
   const char
     *value;
@@ -928,7 +935,7 @@ static MagickBooleanType WriteWEBPImage(const ImageInfo *image_info,
     (void) LogMagickEvent(TraceEvent,GetMagickModule(),"%s",image->filename);
   if ((image->columns > 16383UL) || (image->rows > 16383UL))
     ThrowWriterException(ImageError,"WidthOrHeightExceedsLimit");
-  status=OpenBlob(image_info,image,WriteBinaryBlobMode,&image->exception);
+  status=OpenBlob(image_info,image,WriteBinaryBlobMode,exception);
   if (status == MagickFalse)
     return(status);
   if (WebPConfigInit(&configure) == 0)
@@ -1039,14 +1046,14 @@ static MagickBooleanType WriteWEBPImage(const ImageInfo *image_info,
   if (WebPValidateConfig(&configure) == 0)
     ThrowWriterException(ResourceLimitError,"UnableToEncodeImageFile");
 
-  WriteSingleWEBPImage(image_info,image,&picture,&memory,&image->exception);
+  WriteSingleWEBPImage(image_info,image,&picture,&memory,exception);
 
 #if defined(MAGICKCORE_WEBPMUX_DELEGATE)
   if ((image_info->adjoin != MagickFalse) &&
       (GetPreviousImageInList(image) == (Image *) NULL) &&
       (GetNextImageInList(image) != (Image *) NULL) &&
       (image->iterations != 1))
-    WriteAnimatedWEBPImage(image_info,image,&configure,&writer_info,&image->exception);
+    WriteAnimatedWEBPImage(image_info,image,&configure,&writer_info,exception);
 #endif
 
   webp_status=WebPEncode(&configure,&picture);
@@ -1115,7 +1122,7 @@ static MagickBooleanType WriteWEBPImage(const ImageInfo *image_info,
           break;
         }
       }
-      (void) ThrowMagickException(&image->exception,GetMagickModule(),CorruptImageError,
+      (void) ThrowMagickException(exception,GetMagickModule(),CorruptImageError,
         (char *) message,"`%s'",image->filename);
     }
 #if defined(MAGICKCORE_WEBPMUX_DELEGATE)
@@ -1163,7 +1170,7 @@ static MagickBooleanType WriteWEBPImage(const ImageInfo *image_info,
         mux_error=WebPMuxSetChunk(mux,"XMP",&chunk,0);
       }
     if (mux_error != WEBP_MUX_OK)
-      (void) ThrowMagickException(&image->exception,GetMagickModule(),
+      (void) ThrowMagickException(exception,GetMagickModule(),
         ResourceLimitError,"UnableToEncodeImageFile","`%s'",image->filename);
     if (chunk.size != 0)
       {
